-------------------------------------------------------------- 
--  Subject : Symbolic differentiation ( module )
--  Copyright (c) 1999 by A.Baranovsky
--------------------------------------------------------------

uses global

class TScanner  
  function Init( InitBuff ) constructor                  
  function GetToken() 
end

implementation TScanner
  local Buff, P

  function Init( InitBuff )
    Buff := InitBuff + '#'
    P := 0
  return 0

  function IsDigit( Ch )
  return ( Ch >= '0' ) and ( Ch <= '9' )

  function GetNextChar()
    Inc( P )
    Ch := Substr( Buff, P, 1 )
  return Ch

  function BackUp()
    Dec( P )
  return 0

  function GetToken()
    try
      repeat
        Ch := GetNextChar()
        if Ch = ' ' then
        --------  continue  -----------
        elsif Ch = '#' then
          Token := Ch
          raise Break
        elsif Ch = 'l' then
          Ch := GetNextChar()
          if Ch = 'n' then
            Token := LOG
            raise Break
          else
            raise TFailrue( 'Syntax error' )  
          end  
        elsif IsVar( Ch ) then 
          Token := Ch
          raise Break
        elsif Ch = PLUS then  
          Token := Ch
          raise Break
        elsif Ch = MINUS then 
          Token := Ch
          raise Break
        elsif Ch = MULT then  
          Token := Ch
          raise Break
        elsif Ch = DIV then   
          Token := Ch
          raise Break
        elsif Ch = POW then    
          Token := Ch
          raise Break
        elsif Ch = LOG then     
          Token := Ch
          raise Break
        elsif Ch = SP_L then
          Token := Ch
          raise Break
        elsif Ch = SP_R then
          Token := Ch
          raise Break
        elsif IsDigit( Ch ) then
          S := ''
          while IsDigit( Ch ) do
            S  := S + Ch
            Stack << [ ch, << Stack ]
            Ch := GetNextChar()
          end
          BackUp()
          K := 0
          M := 1 

          Q == Stack
          while not Null( Q ) do
            [[ ch, Q ]] := [ << Q ]
            D := Ord( Ch ) - Ord( '0' )
            K := K + D * M
            M := M * 10
          end
          Token := K
          raise Break
        else
          raise TFailrue( 'Syntax error' )                                 
        end
      until FALSE
    except
      on TFailrue do
        raise
      else
      ----- continue -----------
      end
    end
  return Token
end

class TParser
  function Init( InitScanner ) constructor
  function Parse_EXPRESSION()
  function Parse_TERM()
  function Parse_MULTIPLIER()
  function Parse_FACTOR()
  function Call_SCANNER()
end

implementation TParser

  local CurrToken, Scanner

  function Init( InitScanner )             
    object &Scanner : TScanner
    Scanner == InitScanner

    Call_SCANNER()
  return 0

  function IsPowOper()
  return ( ClassOf( CurrToken ) = ClassString ) and ( CurrToken = POW )

  function IsMultOper()
     if ClassOf( CurrToken ) = ClassString  then
       R := ( CurrToken = MULT ) or ( CurrToken = DIV )
     else
       R := FALSE
     end 
  return R

  function IsAddOper() 
     if ClassOf( CurrToken ) = ClassString  then
       R := ( CurrToken = PLUS ) or ( CurrToken = MINUS )
     else
       R := FALSE
     end 
  return R

  function Parse_EXPRESSION()
    R := Parse_TERM()
    while IsAddOper() do
      OP := CurrToken
      Call_SCANNER()
      R << [ OP, << R, Parse_Term() ] 
    end
  return R

  function Parse_TERM()
    R := Parse_MULTIPLIER()
    while IsMultOper() do
      OP := CurrToken
      Call_SCANNER()
      R << [ OP, << R, Parse_MULTIPLIER() ] 
    end
  return R

  function Parse_MULTIPLIER()
    R := Parse_FACTOR()                
    while IsPowOper() do
      OP := CurrToken
      Call_SCANNER()

      R << [ OP, << R, Parse_FACTOR() ]
    end
  return R


  function Parse_FACTOR()
    if ClassOf( CurrToken ) = ClassInteger then
      R := CurrToken
      Call_SCANNER()
    elsif CurrToken = LOG  then
      Call_SCANNER()

      if not Equal( SP_L, CurrToken ) then
        raise TFailrue( '( expected' )
      end

      Call_SCANNER()
      temp := Parse_EXPRESSION()
      R := [ LOG, temp ]

      if not Equal( SP_R, CurrToken ) then
        raise TFailrue( ') expected' )
      end

      Call_SCANNER()
    elsif CurrToken = SP_L then
      Call_SCANNER()
      R := Parse_EXPRESSION()
      Call_SCANNER()
    elsif CurrToken = MINUS then
      Call_SCANNER()
      R := [ UN, Parse_FACTOR() ]
    elsif IsVar( CurrToken ) then
      R := CurrToken
      Call_SCANNER()
    else
      raise TFailrue( 'A variable or a constant expected' )
    end
  return R

  function Call_SCANNER()
    Destroy( CurrToken )
    CurrToken := Scanner.GetToken()
  return 0
end

return


